import logging
import time

import pytest

from vastdb import util
from vastdb.table import ImportConfig, QueryConfig

log = logging.getLogger(__name__)


@pytest.mark.benchmark
def test_bench(session, bucket_name, parquets_path, crater_path):
    files = [str(parquets_path / f) for f in (parquets_path.glob('**/*.pq'))]

    with session.transaction() as tx:
        b = tx.bucket(bucket_name)
        s = b.create_schema('s1')
        t = util.create_table_from_files(s, 't1', files, config=ImportConfig(import_concurrency=8))
        config = QueryConfig(num_splits=8, num_sub_splits=4)
        s = time.time()
        pa_table = t.select(columns=['sid'], predicate=t['sid'] == 10033007, config=config).read_all()
        e = time.time()
        log.info("'SELECT sid from TABLE WHERE sid = 10033007' returned in %s seconds.", e - s)
        if crater_path:
            with open(f'{crater_path}/bench_results', 'a') as f:
                f.write(f"'SELECT sid FROM TABLE WHERE sid = 10033007' returned in {e - s} seconds")
        assert pa_table.num_rows == 255_075



import logging
import time

import pyarrow as pa
import pytest

from vastdb.table import QueryConfig

logger = logging.getLogger(__name__)


@pytest.mark.ha
def test_ha_query(session, test_bucket_name, schema_name, table_name):
    # runs in parallel to ha scenario
    times_to_query, records_in_table = 50, 100_000_000
    arrow_array = pa.array(range(0, records_in_table), type=pa.int64())

    with session.transaction() as tx:
        config = QueryConfig(num_splits=1, num_sub_splits=1)
        start = time.time()
        for i in range(times_to_query):
            table = tx.bucket(test_bucket_name).schema(schema_name).table(table_name)
            s = time.time()
            res = table.select(config=config).read_all()
            e = time.time()
            logger.info(f'query number {i} took {e - s} sec')
            assert arrow_array == pa.concat_arrays(res.column('a').chunks)
        end = time.time()
        logger.info(f'SDK test took {end - start}')



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# ----------------------------------------------------------------------
# The root Message type
# This union enables us to easily send different message types without
# redundant storage, and in the future we can easily add new message types.
#
# Arrow implementations do not need to implement all of the message types,
# which may include experimental metadata types. For maximum compatibility,
# it is best to send data using RecordBatch
class MessageHeader(object):
    NONE = 0
    Schema = 1
    DictionaryBatch = 2
    RecordBatch = 3
    Tensor = 4
    SparseTensor = 5




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Literal(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Literal()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsLiteral(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Literal
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Literal
    def ImplType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Literal value data; for null literals do not include this field.
    # Literal
    def Impl(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # Type of the literal value. This must match `impl`.
    # Literal
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.flatbuf.Field import Field
            obj = Field()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def Start(builder): builder.StartObject(3)
def LiteralStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddImplType(builder, implType): builder.PrependUint8Slot(0, implType, 0)
def LiteralAddImplType(builder, implType):
    """This method is deprecated. Please switch to AddImplType."""
    return AddImplType(builder, implType)
def AddImpl(builder, impl): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(impl), 0)
def LiteralAddImpl(builder, impl):
    """This method is deprecated. Please switch to AddImpl."""
    return AddImpl(builder, impl)
def AddType(builder, type): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(type), 0)
def LiteralAddType(builder, type):
    """This method is deprecated. Please switch to AddType."""
    return AddType(builder, type)
def End(builder): return builder.EndObject()
def LiteralEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# A concrete bound, which can be an expression or unbounded
class ConcreteBoundImpl(object):
    NONE = 0
    Expression = 1
    Unbounded = 2




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

class JoinKind(object):
    Anti = 0
    Cross = 1
    FullOuter = 2
    Inner = 3
    LeftOuter = 4
    LeftSemi = 5
    RightOuter = 6




from typing import List, Optional

from vastdb.bench.perf_bench.common.constants import (
    LOCAL_FS_DS_PATH,
    NFS_DS_PATH,  # noqa: F401
    S3_DS_PATH,  # noqa: F401
    ParquetCompression,
    VastConnDetails,
)
from vastdb.bench.perf_bench.dataset.schemas import DEFAULT_BARS_COLUMNS
from vastdb.bench.perf_bench.orchestrate.scenario import BenchScenario
from vastdb.bench.perf_bench.orchestrate.scenario_generator import (
    generate_perf_bench_scenarios,
)
from vastdb.bench.perf_bench.query.query import QueryBackend


def build_scenarios(
    base_key: str,
    conn_details: Optional[VastConnDetails] = None,
) -> List[BenchScenario]:
    return generate_perf_bench_scenarios(
        base_key=base_key,
        conn_details=conn_details or VastConnDetails(),
        query_backends=[
            QueryBackend.pyarrow,
            # QueryBackend.vastdb,
        ],
        columns_choices=(DEFAULT_BARS_COLUMNS,),
        universe_choices=(
            "Single",
            "Tiny",
            "SmallSeq",
            "Medium",
            "Medium2",
            "Large",
        ),
        num_bdays=[
            1,  # 1d
            5,  # 1w
            # 22,  # 1m
            65,  # 3m
            # 130,  # 6m
            252,  # 1y
        ],

        # Arrow-specific options
        fs_path_choices=[
            # NFS_DS_PATH,
            LOCAL_FS_DS_PATH,
            # S3_DS_PATH,
        ],
        rowgroup_size_choices=[  # make sure you have previously generated the respective datasets
            # 64 * 1024,
            # 128 * 1024,
            256 * 1024,
            # 512 * 1024,
            # DEFAULT_ROW_GROUP_SIZE,
            # int(1.5 * 1024 * 1024),
        ],
        compression_choices=[
            ParquetCompression.LZ4,
        ],
        arrow_batching_spec_choices=[
            # {"batch_size": 2*2**16, "batch_readahead": 16, "fragment_readahead": 4},
            # {"batch_size": 6*2**16, "batch_readahead": 12, "fragment_readahead": 4},
            # DEFAULT_ARROW_KWARGS,
            {"batch_size": 16 * 2 ** 16, "batch_readahead": 16, "fragment_readahead": 4},
            # {"batch_size": 24 * 2 ** 16, "batch_readahead": 12, "fragment_readahead": 4},
            # {"batch_size": 32*2**16, "batch_readahead": 12, "fragment_readahead": 4},
            # {"batch_size": 64*2**16, "batch_readahead": 12, "fragment_readahead": 4},
            # {"batch_size": 128*2**16, "batch_readahead": 12, "fragment_readahead": 4},
        ],

        # VastDB-specific options
        vdb_num_sub_splits_choices=(
            # 1,  # Default
            # 4,
            8,
            # 16,
        ),
        vdb_num_row_groups_per_sub_split_choices=(
            # 1,
            # 4,
            8,  # Default
        ),
    )



# ruff: noqa: F841, PLW0603

import datetime as dt
import logging
from typing import Any, Dict, Optional, Sequence

import pyarrow as pa

import vastdb.bench.perf_bench.common.log_utils
from vastdb.bench.perf_bench.common import utils as bu
from vastdb.bench.perf_bench.common.constants import VastConnDetails
from vastdb.bench.perf_bench.dataset import secmaster as sm
from vastdb.bench.perf_bench.dataset.schemas import BF
from vastdb.table import QueryConfig

LOG = vastdb.bench.perf_bench.common.log_utils.get_logger(__name__)


# noinspection PyUnusedLocal,PyArgumentList
def query_vastdb(
    universe: Sequence[str],
    columns: Optional[Sequence[str]] = None,
    from_t: Optional[dt.datetime] = None,
    to_t: Optional[dt.datetime] = None,
    path: Optional[str] = None,
    use_sid: bool = True,
    logger: Optional[logging.Logger] = None,
    backend_kwargs: Optional[Dict[str, Any]] = None,
) -> pa.Table:
    # ------------------------------------------------------------
    #    Query via VastDB
    # ------------------------------------------------------------
    kwargs = backend_kwargs or {}
    conn_details: Optional[VastConnDetails] = kwargs.pop("conn_details", None)
    conn_details = conn_details or VastConnDetails()
    ss = bu.get_vastdb_session(
        access=conn_details.access,
        secret=conn_details.secret,
        vastdb_endpoint=conn_details.vastdb_endpoint,
        ssl_verify=conn_details.vastdb_ssl_verify,
    )
    with ss.transaction() as tx:
        b = tx.bucket(conn_details.vastdb_bucket)
        s = b.schema(conn_details.vastdb_schema)
        t = s.table(conn_details.vastdb_table)

        if use_sid:
            fld = BF.sid.value
            sid_uni = sorted(sm.to_sid(s) for s in universe)
            filters = t[fld].isin(sid_uni)
        else:
            fld = BF.ticker.value
            filters = t[fld].isin(sorted(universe))

        if from_t and to_t:
            # noinspection PyTypedDict
            filters = filters & (t["ts"] >= from_t) & (t["ts"] < to_t)
        elif from_t:
            filters = filters & (t["ts"] >= from_t)
        elif to_t:
            filters.append(t["ts"] < to_t)

        # Cleanup the kwargs to be passed to the arrow scanner
        kwargs.pop("arrow_kwargs", None)
        if kwargs.pop("filter", None):
            raise ValueError("Can't use filter with VastDB query")

        # Perform the query
        # noinspection PyTypeChecker
        config = QueryConfig(**kwargs)
        endpoint = conn_details.vastdb_endpoint
        config.num_splits = 32
        config.num_sub_splits = 8
        config.data_endpoints = [endpoint] * config.num_splits
        table = pa.Table.from_batches(
            batches=t.select(predicate=filters, columns=columns, config=config)
        )
    return table



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ListTablesResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ListTablesResponse()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsListTablesResponse(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # ListTablesResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ListTablesResponse
    def BucketName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ListTablesResponse
    def SchemaName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ListTablesResponse
    def Tables(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vast_flatbuf.tabular.ObjectDetails import ObjectDetails
            obj = ObjectDetails()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ListTablesResponse
    def TablesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ListTablesResponse
    def TablesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def Start(builder): builder.StartObject(3)
def ListTablesResponseStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddBucketName(builder, bucketName): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(bucketName), 0)
def ListTablesResponseAddBucketName(builder, bucketName):
    """This method is deprecated. Please switch to AddBucketName."""
    return AddBucketName(builder, bucketName)
def AddSchemaName(builder, schemaName): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(schemaName), 0)
def ListTablesResponseAddSchemaName(builder, schemaName):
    """This method is deprecated. Please switch to AddSchemaName."""
    return AddSchemaName(builder, schemaName)
def AddTables(builder, tables): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(tables), 0)
def ListTablesResponseAddTables(builder, tables):
    """This method is deprecated. Please switch to AddTables."""
    return AddTables(builder, tables)
def StartTablesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ListTablesResponseStartTablesVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartTablesVector(builder, numElems)
def End(builder): return builder.EndObject()
def ListTablesResponseEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


import functools
import itertools
import operator

import pyarrow as pa
import pyarrow.compute as pc
import pytest

from .util import prepare_data


def test_nested_select(session, clean_bucket_name):
    columns = pa.schema([
        ('l', pa.list_(pa.int8())),
        ('m', pa.map_(pa.utf8(), pa.float64())),
        ('s', pa.struct([('x', pa.int16()), ('y', pa.int32())])),
    ])
    expected = pa.table(schema=columns, data=[
        [[1], [], [2, 3], None],
        [None, {'a': 2.5}, {'b': 0.25, 'c': 0.025}, {}],
        [{'x': 1, 'y': None}, None, {'x': 2, 'y': 3}, {'x': None, 'y': 4}],
    ])

    with prepare_data(session, clean_bucket_name, 's', 't', expected) as t:
        actual = t.select().read_all()
        assert actual == expected

        names = [f.name for f in columns]
        for n in range(len(names) + 1):
            for cols in itertools.permutations(names, n):
                actual = t.select(columns=cols).read_all()
                assert actual == expected.select(cols)


def test_nested_filter(session, clean_bucket_name):
    columns = pa.schema([
        ('x', pa.int64()),
        ('l', pa.list_(pa.int8())),
        ('y', pa.int64()),
        ('m', pa.map_(pa.utf8(), pa.float64())),
        ('z', pa.int64()),
        ('s', pa.struct([('x', pa.int16()), ('y', pa.int32())])),
        ('w', pa.int64()),
    ])
    expected = pa.table(schema=columns, data=[
        [1, 2, 3, None],
        [[1], [], [2, 3], None],
        [1, 2, None, 3],
        [None, {'a': 2.5}, {'b': 0.25, 'c': 0.025}, {}],
        [1, None, 2, 3],
        [{'x': 1, 'y': None}, None, {'x': 2, 'y': 3}, {'x': None, 'y': 4}],
        [None, 1, 2, 3],
    ])

    with prepare_data(session, clean_bucket_name, 's', 't', expected) as t:
        actual = t.select().read_all()
        assert actual == expected

        names = list('xyzw')
        for n in range(1, len(names) + 1):
            for cols in itertools.permutations(names, n):
                ibis_predicate = functools.reduce(
                    operator.and_,
                    (t[col] > 2 for col in cols))
                actual = t.select(predicate=ibis_predicate).read_all()

                arrow_predicate = functools.reduce(
                    operator.and_,
                    (pc.field(col) > 2 for col in cols))
                assert actual == expected.filter(arrow_predicate)


def test_nested_unsupported_filter(session, clean_bucket_name):
    columns = pa.schema([
        ('x', pa.int64()),
        ('l', pa.list_(pa.int8())),
        ('y', pa.int64()),
        ('m', pa.map_(pa.utf8(), pa.float64())),
        ('z', pa.int64()),
        ('s', pa.struct([('x', pa.int16()), ('y', pa.int32())])),
        ('w', pa.int64()),
    ])
    expected = pa.table(schema=columns, data=[
        [1, 2, 3, None],
        [[1], [], [2, 3], None],
        [1, 2, None, 3],
        [None, {'a': 2.5}, {'b': 0.25, 'c': 0.025}, {}],
        [1, None, 2, 3],
        [{'x': 1, 'y': None}, None, {'x': 2, 'y': 3}, {'x': None, 'y': 4}],
        [None, 1, 2, 3],
    ])

    with prepare_data(session, clean_bucket_name, 's', 't', expected) as t:

        with pytest.raises(NotImplementedError):
            list(t.select(predicate=(t['l'].isnull())))

        with pytest.raises(NotImplementedError):
            list(t.select(predicate=(t['m'].isnull())))

        with pytest.raises(NotImplementedError):
            list(t.select(predicate=(t['s'].isnull())))


def test_nested_subfields_predicate_pushdown(session, clean_bucket_name):
    columns = pa.schema([
        ('x', pa.int64()),
        ('l', pa.list_(pa.int8())),
        ('y', pa.int64()),
        ('m', pa.map_(pa.utf8(), pa.float64())),
        ('z', pa.int64()),
        ('s', pa.struct([
            ('x', pa.int16()),
            ('y', pa.int32()),
            ('q', pa.struct([
                ('q1', pa.utf8()),
                ('q2', pa.float32())
            ]))
        ])),
        ('w', pa.int64()),
    ])
    expected = pa.table(schema=columns, data=[
        [1, 2, 3, None],
        [[1], [], [2, 3], None],
        [1, 2, None, 3],
        [None, {'a': 2.5}, {'b': 0.25, 'c': 0.025}, {}],
        [1, None, 2, 3],
        [
            {'x': 1, 'y': None, 'q': {'q1': 'AAA', 'q2': 1.0}},
            None,
            {'x': 2, 'y': 3, 'q': {'q1': 'B', 'q2': 2.0}},
            {'x': None, 'y': 4, 'q': {'q1': 'CC', 'q2': 2.0}}],
        [None, 1, 2, 3],
    ])

    with prepare_data(session, clean_bucket_name, 's', 't', expected) as t:

        assert t.select(predicate=(t['s']['x'] == 1)).read_all() == expected.take([0])
        assert t.select(predicate=(t['s']['y'].isnull())).read_all() == expected.take([0, 1])
        assert t.select(predicate=(t['s']['q']['q1'] == 'AAA')).read_all() == expected.take([0])
        assert t.select(predicate=(t['s']['q']['q1'] < 'B')).read_all() == expected.take([0])
        assert t.select(predicate=(t['s']['q']['q1'] <= 'B')).read_all() == expected.take([0, 2])
        assert t.select(predicate=(t['s']['q']['q2'] == 1.0)).read_all() == expected.take([0])

        assert t.select(predicate=(t['s']['q']['q1'].isnull())).read_all() == expected.take([1])
        assert t.select(predicate=(t['s']['q']['q2'].isnull())).read_all() == expected.take([1])

        assert t.select(predicate=(t['s']['x'] == 2)).read_all() == expected.take([2])
        assert t.select(predicate=(t['s']['y'] == 3)).read_all() == expected.take([2])
        assert t.select(predicate=(t['s']['q']['q1'] == 'B')).read_all() == expected.take([2])
        assert t.select(predicate=(t['s']['q']['q2'] == 2.0)).read_all() == expected.take([2, 3])

        assert t.select(predicate=(t['s']['x'].isnull())).read_all() == expected.take([1, 3])
        assert t.select(predicate=(t['s']['y'] == 4)).read_all() == expected.take([3])
        assert t.select(predicate=(t['s']['q']['q1'] == 'CC')).read_all() == expected.take([3])
        assert t.select(predicate=(t['s']['q']['q1'] > 'B')).read_all() == expected.take([3])
        assert t.select(predicate=(t['s']['q']['q1'] >= 'B')).read_all() == expected.take([2, 3])

        assert t.select(predicate=(t['s']['x'] == 1) | (t['s']['x'] == 2)).read_all() == expected.take([0, 2])
        assert t.select(predicate=(t['s']['x'].isnull()) & (t['s']['y'].isnull())).read_all() == expected.take([1])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class AlterColumnRequest(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = AlterColumnRequest()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAlterColumnRequest(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # AlterColumnRequest
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AlterColumnRequest
    def Properties(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # AlterColumnRequest
    def Stats(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def Start(builder): builder.StartObject(2)
def AlterColumnRequestStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddProperties(builder, properties): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(properties), 0)
def AlterColumnRequestAddProperties(builder, properties):
    """This method is deprecated. Please switch to AddProperties."""
    return AddProperties(builder, properties)
def AddStats(builder, stats): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(stats), 0)
def AlterColumnRequestAddStats(builder, stats):
    """This method is deprecated. Please switch to AddStats."""
    return AddStats(builder, stats)
def End(builder): return builder.EndObject()
def AlterColumnRequestEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# A union of possible dereference operations
class Deref(object):
    NONE = 0
    # Access a value for a given map key
    MapKey = 1
    # Access the value at a struct field
    StructField = 2
    # Access the element at a given index in an array
    ArraySubscript = 3
    # Access a range of elements in an array
    ArraySlice = 4
    # Access a field of a relation
    FieldIndex = 5




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

class LiteralImpl(object):
    NONE = 0
    BooleanLiteral = 1
    Int8Literal = 2
    Int16Literal = 3
    Int32Literal = 4
    Int64Literal = 5
    UInt8Literal = 6
    UInt16Literal = 7
    UInt32Literal = 8
    UInt64Literal = 9
    DateLiteral = 10
    TimeLiteral = 11
    TimestampLiteral = 12
    IntervalLiteral = 13
    DurationLiteral = 14
    DecimalLiteral = 15
    Float16Literal = 16
    Float32Literal = 17
    Float64Literal = 18
    ListLiteral = 19
    StructLiteral = 20
    MapLiteral = 21
    StringLiteral = 22
    BinaryLiteral = 23
    FixedSizeBinaryLiteral = 24




from functools import lru_cache
from types import MappingProxyType
from typing import Dict, Mapping, Union

import numpy as np
import pyarrow as pa
from pandas.core.dtypes.base import ExtensionDtype

from vastdb.bench.perf_bench.common.types import StrEnum

PandasDTypeT = Union[ExtensionDtype, np.dtype]


class StockBarField(StrEnum):
    sid = "sid"
    ts = "ts"
    ticker = "ticker"

    # Ask
    ask_open = "ask_open"
    ask_high = "ask_high"
    ask_low = "ask_low"
    ask_close = "ask_close"
    ask_qty = "ask_qty"

    # Bid
    bid_open = "bid_open"
    bid_high = "bid_high"
    bid_low = "bid_low"
    bid_close = "bid_close"
    bid_qty = "bid_qty"

    # Trades
    trade_open = "trade_open"
    trade_high = "trade_high"
    trade_low = "trade_low"
    trade_close = "trade_close"
    trade_volume = "trade_volume"

    # VWAP
    vwap = "vwap"

    @property
    def pa_type(self) -> pa.DataType:
        return _get_field_pyarrow_types()[self]

    @property
    def pd_type(self) -> PandasDTypeT:
        return self.pa_type.to_pandas_dtype()


BF = StockBarField


@lru_cache
def _get_field_pyarrow_types() -> Mapping[str, pa.DataType]:
    return MappingProxyType(
        {
            BF.sid.value: pa.int64(),
            BF.ts.value: pa.timestamp(unit="ns"),
            BF.ticker.value: pa.string(),
            BF.ask_open.value: pa.float64(),
            BF.ask_high.value: pa.float64(),
            BF.ask_low.value: pa.float64(),
            BF.ask_close.value: pa.float64(),
            BF.ask_qty.value: pa.int64(),
            BF.bid_open.value: pa.float64(),
            BF.bid_high.value: pa.float64(),
            BF.bid_low.value: pa.float64(),
            BF.bid_close.value: pa.float64(),
            BF.bid_qty.value: pa.int64(),
            BF.trade_open.value: pa.float64(),
            BF.trade_high.value: pa.float64(),
            BF.trade_low.value: pa.float64(),
            BF.trade_close.value: pa.float64(),
            BF.trade_volume.value: pa.int64(),
            BF.vwap.value: pa.float64(),
        }
    )


DEFAULT_BARS_COLUMNS = (
    BF.sid.value,
    BF.ts.value,
    BF.ask_open.value,
    BF.ask_close.value,
    BF.bid_open.value,
    BF.bid_close.value,
    BF.bid_qty.value,
)

# noinspection PyUnresolvedReferences
StockBarsArrowSchema: pa.Schema = pa.schema(
    (fld.value, fld.pa_type) for fld in StockBarField
)
# noinspection PyUnresolvedReferences
StockBarsPandasSchema: Dict[str, PandasDTypeT] = {
    fld.value: fld.pd_type for fld in StockBarField
}

BarsSortFields = (BF.sid.value, BF.ts.value)



import logging
import xml.etree.ElementTree
from dataclasses import dataclass
from enum import Enum

import requests


class HttpStatus(Enum):
    SUCCESS = 200
    BAD_REQUEST = 400
    FOBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    REQUEST_TIMEOUT = 408
    CONFLICT = 409
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    SERVICE_UNAVAILABLE = 503


log = logging.getLogger(__name__)


@dataclass
class HttpError(Exception):
    code: str
    message: str
    method: str
    url: str
    status: int  # HTTP status
    headers: requests.structures.CaseInsensitiveDict  # HTTP response headers

    def __post_init__(self):
        self.args = [vars(self)]


class NotFound(HttpError):
    pass


class Forbidden(HttpError):
    pass


class BadRequest(HttpError):
    pass


class MethodNotAllowed(HttpError):
    pass


class RequestTimeout(HttpError):
    pass


class Conflict(HttpError):
    pass


class InternalServerError(HttpError):
    pass


class NotImplemented(HttpError):
    pass


class ServiceUnavailable(HttpError):
    pass


class Slowdown(ServiceUnavailable):
    pass


class UnexpectedError(HttpError):
    pass


@dataclass
class ImportFilesError(Exception):
    message: str
    error_dict: dict


class InvalidArgument(Exception):
    pass


class TooLargeRequest(InvalidArgument):
    pass


class TooWideRow(TooLargeRequest):
    pass


class Missing(Exception):
    pass


class MissingTransaction(Missing):
    pass


class MissingRowIdColumn(Missing):
    pass


class NotSupported(Exception):
    pass


@dataclass
class MissingBucket(Missing):
    bucket: str


@dataclass
class MissingSnapshot(Missing):
    bucket: str
    snapshot: str


@dataclass
class MissingSchema(Missing):
    bucket: str
    schema: str


@dataclass
class MissingTable(Missing):
    bucket: str
    schema: str
    table: str


@dataclass
class MissingProjection(Missing):
    bucket: str
    schema: str
    table: str
    projection: str


class Exists(Exception):
    pass


@dataclass
class SchemaExists(Exists):
    bucket: str
    schema: str


@dataclass
class TableExists(Exists):
    bucket: str
    schema: str
    table: str


@dataclass
class NotSupportedCommand(NotSupported):
    bucket: str
    schema: str
    table: str


@dataclass
class NotSupportedVersion(NotSupported):
    err_msg: str
    version: str


@dataclass
class ConnectionError(Exception):
    cause: Exception
    may_retry: bool


def handle_unavailable(**kwargs):
    if kwargs['code'] == 'SlowDown':
        raise Slowdown(**kwargs)
    raise ServiceUnavailable(**kwargs)


ERROR_TYPES_MAP = {
    HttpStatus.BAD_REQUEST: BadRequest,
    HttpStatus.FOBIDDEN: Forbidden,
    HttpStatus.NOT_FOUND: NotFound,
    HttpStatus.METHOD_NOT_ALLOWED: MethodNotAllowed,
    HttpStatus.REQUEST_TIMEOUT: RequestTimeout,
    HttpStatus.CONFLICT: Conflict,
    HttpStatus.INTERNAL_SERVER_ERROR: InternalServerError,
    HttpStatus.NOT_IMPLEMENTED: NotImplemented,
    HttpStatus.SERVICE_UNAVAILABLE: handle_unavailable,
}


def from_response(res: requests.Response):
    if res.status_code == HttpStatus.SUCCESS.value:
        return None

    log.debug("response: url='%s', code=%s, headers=%s, body='%s'", res.request.url, res.status_code, res.headers, res.text)
    # try to parse S3 XML response for the error details:
    code_str = None
    message_str = None
    if res.text:
        try:
            root = xml.etree.ElementTree.fromstring(res.text)
            code = root.find('Code')
            code_str = code.text if code is not None else None
            message = root.find('Message')
            message_str = message.text if message is not None else None
        except xml.etree.ElementTree.ParseError:
            log.debug("invalid XML: %r", res.text)

    kwargs = dict(
        code=code_str,
        message=message_str,
        method=res.request.method,
        url=res.request.url,
        status=res.status_code,
        headers=res.headers,
    )
    log.warning("RPC failed: %s", kwargs)
    status = HttpStatus(res.status_code)
    error_type = ERROR_TYPES_MAP.get(status, UnexpectedError)
    return error_type(**kwargs)  # type: ignore



import os
import sqlite3
from pathlib import Path

import boto3
import pytest

import vastdb


def pytest_addoption(parser):
    parser.addoption("--tabular-bucket-name", help="Name of the S3 bucket with Tabular enabled", default="vastdb")
    parser.addoption("--tabular-access-key", help="Access key with Tabular permissions (AWS_ACCESS_KEY_ID)",
                     default=os.environ.get("AWS_ACCESS_KEY_ID", None))
    parser.addoption("--tabular-secret-key", help="Secret key with Tabular permissions (AWS_SECRET_ACCESS_KEY)",
                     default=os.environ.get("AWS_SECRET_ACCESS_KEY", None))
    parser.addoption("--tabular-endpoint-url", help="Tabular server endpoint", default=[], action="append")
    parser.addoption("--data-path", help="Data files location", default=None)
    parser.addoption("--crater-path", help="Save benchmark results in a dedicated location", default=None)
    parser.addoption("--schema-name", help="Name of schema for the test to operate on", default=None)
    parser.addoption("--table-name", help="Name of table for the test to operate on", default=None)
    parser.addoption("--num-workers", help="Number of concurrent workers", default=1)


@pytest.fixture(scope="session")
def session_kwargs(request, tabular_endpoint_urls):
    return dict(
        access=request.config.getoption("--tabular-access-key"),
        secret=request.config.getoption("--tabular-secret-key"),
        endpoint=tabular_endpoint_urls[0],
    )


@pytest.fixture(scope="session")
def session(session_kwargs):
    return vastdb.connect(**session_kwargs)


@pytest.fixture(scope="session")
def num_workers(request):
    return int(request.config.getoption("--num-workers"))


@pytest.fixture(scope="session")
def test_bucket_name(request):
    return request.config.getoption("--tabular-bucket-name")


@pytest.fixture(scope="session")
def tabular_endpoint_urls(request):
    return request.config.getoption("--tabular-endpoint-url") or ["http://localhost:9090"]


def iter_schemas(s):
    """Recusively scan all schemas."""
    children = s.schemas()
    for c in children:
        yield from iter_schemas(c)
    yield s


@pytest.fixture(scope="function")
def clean_bucket_name(request, test_bucket_name, session):
    with session.transaction() as tx:
        b = tx.bucket(test_bucket_name)
        for top_schema in b.schemas():
            for s in iter_schemas(top_schema):
                for t in s.tables():
                    t.drop()
                s.drop()
    return test_bucket_name


@pytest.fixture(scope="session")
def s3(request, tabular_endpoint_urls):
    return boto3.client(
        's3',
        aws_access_key_id=request.config.getoption("--tabular-access-key"),
        aws_secret_access_key=request.config.getoption("--tabular-secret-key"),
        endpoint_url=tabular_endpoint_urls[0])


@pytest.fixture(scope="function")
def parquets_path(request):
    return Path(request.config.getoption("--data-path"))


@pytest.fixture(scope="function")
def crater_path(request):
    return request.config.getoption("--crater-path")


@pytest.fixture(scope="function")
def schema_name(request):
    return request.config.getoption("--schema-name")


@pytest.fixture(scope="function")
def table_name(request):
    return request.config.getoption("--table-name")


@pytest.fixture(scope="function")
def perf_metrics_db(crater_path):
    return sqlite3.connect(f"{crater_path}/metrics.sqlite" if crater_path else ":memory:")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Limit operation
class Limit(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Limit()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsLimit(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Limit
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # An identifiier for the relation. The identifier should be unique over the
    # entire plan. Optional.
    # Limit
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.RelId import RelId
            obj = RelId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Child relation
    # Limit
    def Rel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.Relation import Relation
            obj = Relation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Starting index of rows
    # Limit
    def Offset(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # The maximum number of rows of output.
    # Limit
    def Count(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(4)
def LimitStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddId(builder, id): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def LimitAddId(builder, id):
    """This method is deprecated. Please switch to AddId."""
    return AddId(builder, id)
def AddRel(builder, rel): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(rel), 0)
def LimitAddRel(builder, rel):
    """This method is deprecated. Please switch to AddRel."""
    return AddRel(builder, rel)
def AddOffset(builder, offset): builder.PrependUint32Slot(2, offset, 0)
def LimitAddOffset(builder, offset):
    """This method is deprecated. Please switch to AddOffset."""
    return AddOffset(builder, offset)
def AddCount(builder, count): builder.PrependUint32Slot(3, count, 0)
def LimitAddCount(builder, count):
    """This method is deprecated. Please switch to AddCount."""
    return AddCount(builder, count)
def End(builder): return builder.EndObject()
def LimitEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


import pyarrow as pa
import pytest

from .. import errors, util


def test_slices():
    ROWS = 1 << 20
    t = pa.table({"x": range(ROWS), "y": [i / 1000 for i in range(ROWS)]})

    chunks = list(util.iter_serialized_slices(t))
    assert len(chunks) > 1
    sizes = [len(c) for c in chunks]

    assert max(sizes) < util.MAX_RECORD_BATCH_SLICE_SIZE
    assert t == pa.Table.from_batches(_parse(chunks))

    chunks = list(util.iter_serialized_slices(t, 1000))
    assert len(chunks) > 1
    sizes = [len(c) for c in chunks]

    assert max(sizes) < util.MAX_RECORD_BATCH_SLICE_SIZE
    assert t == pa.Table.from_batches(_parse(chunks))


def test_wide_row():
    cols = [pa.field(f"x{i}", pa.utf8()) for i in range(1000)]
    values = [['a' * 10000]] * len(cols)
    t = pa.table(values, schema=pa.schema(cols))
    assert len(t) == 1

    with pytest.raises(errors.TooWideRow):
        list(util.iter_serialized_slices(t))


def test_expand_ip_ranges():
    endpoints = ["http://172.19.101.1-3"]
    expected = ["http://172.19.101.1", "http://172.19.101.2", "http://172.19.101.3"]
    assert util.expand_ip_ranges(endpoints) == expected


def _parse(bufs):
    for buf in bufs:
        with pa.ipc.open_stream(buf) as reader:
            yield from reader


def test_prefix():
    assert util.prefix_to_range('a') == (b'a', b'b')
    assert util.prefix_to_range('abc') == (b'abc', b'abd')
    assert util.prefix_to_range('abc\x00') == (b'abc\x00', b'abc\x01')
    assert util.prefix_to_range('abc\x7f') == (b'abc\x7f', b'abc\x80')
    assert util.prefix_to_range('/a/b/c') == (b'/a/b/c', b'/a/b/d')
    assert util.prefix_to_range('/123α') == (b'/123\xce\xb1', b'/123\xce\xb2')
    assert util.prefix_to_range('/123αA') == (b'/123\xce\xb1A', b'/123\xce\xb1B')
    assert util.prefix_to_range('\U0010ffff') == (b'\xf4\x8f\xbf\xbf', b'\xf4\x8f\xbf\xc0')  # max unicode codepoint
    with pytest.raises(AssertionError):
        util.prefix_to_range('')



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# ----------------------------------------------------------------------
# A Buffer represents a single contiguous memory segment
class Buffer(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 16

    # Buffer
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The relative offset into the shared memory page where the bytes for this
    # buffer starts
    # Buffer
    def Offset(self): return self._tab.Get(flatbuffers.number_types.Int64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # The absolute length (in bytes) of the memory buffer. The memory is found
    # from offset (inclusive) to offset + length (non-inclusive). When building
    # messages using the encapsulated IPC message, padding bytes may be written
    # after a buffer, but such padding bytes do not need to be accounted for in
    # the size here.
    # Buffer
    def Length(self): return self._tab.Get(flatbuffers.number_types.Int64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateBuffer(builder, offset, length):
    builder.Prep(8, 16)
    builder.PrependInt64(length)
    builder.PrependInt64(offset)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Int(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Int()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsInt(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Int
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Int
    def BitWidth(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Int
    def IsSigned(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def Start(builder): builder.StartObject(2)
def IntStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddBitWidth(builder, bitWidth): builder.PrependInt32Slot(0, bitWidth, 0)
def IntAddBitWidth(builder, bitWidth):
    """This method is deprecated. Please switch to AddBitWidth."""
    return AddBitWidth(builder, bitWidth)
def AddIsSigned(builder, isSigned): builder.PrependBoolSlot(1, isSigned, 0)
def IntAddIsSigned(builder, isSigned):
    """This method is deprecated. Please switch to AddIsSigned."""
    return AddIsSigned(builder, isSigned)
def End(builder): return builder.EndObject()
def IntEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# used by `ListSchemas` and `ListTables` API
class ObjectDetails(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ObjectDetails()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsObjectDetails(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # ObjectDetails
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ObjectDetails
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ObjectDetails
    def Properties(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ObjectDetails
    def Handle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ObjectDetails
    def NumRows(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # ObjectDetails
    def SizeInBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # ObjectDetails
    def NumPartitions(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # ObjectDetails
    def SortingKeyEnabled(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def Start(builder): builder.StartObject(7)
def ObjectDetailsStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddName(builder, name): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def ObjectDetailsAddName(builder, name):
    """This method is deprecated. Please switch to AddName."""
    return AddName(builder, name)
def AddProperties(builder, properties): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(properties), 0)
def ObjectDetailsAddProperties(builder, properties):
    """This method is deprecated. Please switch to AddProperties."""
    return AddProperties(builder, properties)
def AddHandle(builder, handle): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(handle), 0)
def ObjectDetailsAddHandle(builder, handle):
    """This method is deprecated. Please switch to AddHandle."""
    return AddHandle(builder, handle)
def AddNumRows(builder, numRows): builder.PrependUint64Slot(3, numRows, 0)
def ObjectDetailsAddNumRows(builder, numRows):
    """This method is deprecated. Please switch to AddNumRows."""
    return AddNumRows(builder, numRows)
def AddSizeInBytes(builder, sizeInBytes): builder.PrependUint64Slot(4, sizeInBytes, 0)
def ObjectDetailsAddSizeInBytes(builder, sizeInBytes):
    """This method is deprecated. Please switch to AddSizeInBytes."""
    return AddSizeInBytes(builder, sizeInBytes)
def AddNumPartitions(builder, numPartitions): builder.PrependUint64Slot(5, numPartitions, 0)
def ObjectDetailsAddNumPartitions(builder, numPartitions):
    """This method is deprecated. Please switch to AddNumPartitions."""
    return AddNumPartitions(builder, numPartitions)
def AddSortingKeyEnabled(builder, sortingKeyEnabled): builder.PrependBoolSlot(6, sortingKeyEnabled, 0)
def ObjectDetailsAddSortingKeyEnabled(builder, sortingKeyEnabled):
    """This method is deprecated. Please switch to AddSortingKeyEnabled."""
    return AddSortingKeyEnabled(builder, sortingKeyEnabled)
def End(builder): return builder.EndObject()
def ObjectDetailsEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

class SparseTensorIndex(object):
    NONE = 0
    SparseTensorIndexCOO = 1
    SparseMatrixIndexCSX = 2
    SparseTensorIndexCSF = 3




# ruff: noqa: F841

import datetime as dt
from functools import reduce
from typing import Optional, Sequence

import pyarrow as pa
from pyarrow import dataset as ds

from vastdb.bench.perf_bench.common import utils as bu
from vastdb.bench.perf_bench.dataset import secmaster as sm
from vastdb.bench.perf_bench.dataset.schemas import BF


def build_arrow_filter(
    universe: Sequence[str],
    from_t: Optional[dt.datetime] = None,
    to_t: Optional[dt.datetime] = None,
    use_sid: bool = True,
) -> ds.Expression:
    # Build the filters
    y_m_filter = from_t_filter = to_t_filter = None
    from_t, to_t = bu.to_ts(from_t, normalize=False), bu.to_ts(to_t, normalize=False)
    if from_t or to_t:
        all_dates = bu.get_dates_range(from_t, to_t, inclusive="left")
        years = sorted({d.year for d in all_dates})

        y_m_filter = ds.field("year").isin(years) & ds.field("date").isin(
            tuple(d.strftime("%Y%m%d") for d in all_dates)
        )

        if from_t and from_t.time() != dt.time.min:
            from_t_filter = ds.field("ts") >= pa.scalar(from_t)
        if to_t and to_t.time() != dt.time.min:
            to_t_filter = ds.field("ts") < pa.scalar(to_t)

    if use_sid:
        universe = [sm.to_sid(s) for s in universe]  # type: ignore[misc]
        fld = BF.sid.value
    else:
        fld = BF.ticker.value
    universe = sorted(universe)
    if len(universe) <= 16:
        tickers_filter = reduce(
            lambda x, y: x | y, [ds.field(fld) == s for s in universe]
        )
    else:
        tickers_filter = ds.field(fld).isin(universe)

    ds_filter = reduce(
        lambda x, y: x & y,
        [
            flt
            for flt in (y_m_filter, tickers_filter, from_t_filter, to_t_filter)
            if flt is not None
        ],
    )

    return ds_filter



import importlib.util
import logging
import os
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from types import MappingProxyType, ModuleType
from typing import (
    Any,
    Dict,
    Generator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import exchange_calendars as xcal
import fsspec
import pandas as pd
from fsspec import AbstractFileSystem

# noinspection PyProtectedMember
from pandas._typing import IntervalClosedType
from pyarrow import dataset as ds

from vastdb.bench.perf_bench.common.constants import (
    DEFAULT_ACCESS_KEY,
    DEFAULT_END_T,
    DEFAULT_ROW_GROUP_SIZE,
    DEFAULT_S3_HOST,
    DEFAULT_SECRET_KEY,
    DEFAULT_START_T,
    DFAULT_PARQUET_COMPRESSION,
    VASTDB_ENDPOINT,
    ParquetCompression,
)
from vastdb.bench.perf_bench.common.log_utils import get_logger
from vastdb.bench.perf_bench.common.types import DateLikeT, PathLikeT
from vastdb.session import Session

_FS_MAP = MappingProxyType({"/mnt": "nfs", "/": "fs", "s3://": "s3", "": "vastdb"})


def config_ipython():
    try:
        # noinspection PyUnresolvedReferences
        _ = get_ipython
        pd.set_option("display.float_format", lambda x: "%.6f" % x)
        pd.set_option("display.max_rows", 50)
        pd.set_option("display.max_columns", 50)
        pd.set_option("display.width", 1000)
    except NameError:
        ...


# If in interactive mode, make it pretty
config_ipython()


def getenv_flag(name: str, default: bool = False) -> bool:
    return os.getenv(name, str(default)).lower() in ("1", "true", "yes")


def to_ts(d: Optional[DateLikeT], normalize: bool = True) -> Optional[pd.Timestamp]:
    if not d:
        return None
    ret_d: pd.Timestamp = d if isinstance(d, pd.Timestamp) else pd.Timestamp(d)
    return ret_d.normalize() if normalize else ret_d


def _default_s3_kwargs(
    aws_access_key_id: str = DEFAULT_ACCESS_KEY,
    aws_secret_access_key: str = DEFAULT_SECRET_KEY,
    host: str = DEFAULT_S3_HOST,
    config: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Dict[str, Any]:
    return {
        "endpoint_url": f"http://{host}",
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
        "use_ssl": False,
        "verify": False,
        "config": {
            "signature_version": "s3v4",
            "s3": {"addressing_style": "path"},
            **(config or {}),
        },
        "region_name": "us-east-1",
        **kwargs,
    }


def get_vastdb_session(
    access: str = DEFAULT_ACCESS_KEY,
    secret: str = DEFAULT_SECRET_KEY,
    vastdb_endpoint: str = VASTDB_ENDPOINT,
    ssl_verify: bool = True,
) -> Session:
    return Session(
        endpoint=vastdb_endpoint, access=access, secret=secret, ssl_verify=ssl_verify
    )


def get_filesystem(
    path: PathLikeT,
    botocore_client_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Tuple[str, AbstractFileSystem]:
    path = str(path)
    if path.startswith("s3://"):
        botocore_client_kwargs = _default_s3_kwargs(
            **(botocore_client_kwargs or {}),
            config=kwargs.pop("config_kwargs", None),
        )
        fs = fsspec.filesystem(
            protocol="s3",
            client_kwargs=botocore_client_kwargs,
            **kwargs,
        )
        path = path[5:]  # remove the s3:// prefix
    else:
        fs = fsspec.filesystem(protocol="file", **kwargs)
    return path, fs


# noinspection PyShadowingBuiltins
def get_parquet_dataset(
    path: PathLikeT,
    filesystem: Optional[AbstractFileSystem] = None,
    fs_kwargs: Optional[Dict[str, Any]] = None,
    format: Optional[str] = "parquet",  # noqa: A002
    partitioning: Optional[str] = "hive",
    **kwargs,
) -> ds.Dataset:
    path, fs = filesystem or get_filesystem(path=path, **(fs_kwargs or {}))
    return ds.dataset(
        source=path,
        format=format,
        partitioning=partitioning,
        filesystem=fs,
        **kwargs,
    )


def get_parquet_dataset_root(
    base_dir: PathLikeT,
    row_group_size: int = DEFAULT_ROW_GROUP_SIZE,
    compression: Union[str, ParquetCompression] = DFAULT_PARQUET_COMPRESSION,
) -> Path:
    return Path(str(base_dir)) / f"rg_{row_group_size!s}_c_{ParquetCompression[compression]}"


@contextmanager
def time_me(logger: Optional[logging.Logger] = None):
    logger = logger or get_logger(__name__)
    start_t = time.time_ns()
    yield  # Yield control back to the context block
    total_t_sec = (time.time_ns() - start_t) / 1e9
    logger.debug(f"Execution time: {total_t_sec:.2f} s")


S = TypeVar("S", bound=Sequence)


def chunk_sequence(input_seq: S, chunk_size: int) -> Generator[Sequence, None, None]:
    for i in range(0, len(input_seq), chunk_size):
        yield input_seq[i: i + chunk_size]


def get_dates_range(
    from_t: Optional[DateLikeT] = DEFAULT_START_T,
    to_t: Optional[DateLikeT] = DEFAULT_END_T,
    only_bdays: bool = True,
    inclusive: IntervalClosedType = "both",
) -> pd.DatetimeIndex:
    fun = pd.bdate_range if only_bdays else pd.date_range
    return fun(  # type: ignore[operator]
        start=to_ts(from_t),
        end=to_ts(to_t),
        inclusive=inclusive,
    )


def infer_fs_type(path: str) -> str:
    return next(k for m, k in _FS_MAP.items() if str(path).lower().startswith(m))


def get_session_minutes(
    date: DateLikeT,
    exchange: str = "XNYS",
) -> pd.DatetimeIndex:
    ecal = xcal.get_calendar(exchange)
    date = pd.Timestamp(date).normalize()
    sess = ecal.date_to_session(date=date, direction="next")
    sess_minutes = ecal.session_minutes(sess)
    return sess_minutes


def load_module_from_path(module_file_path: PathLikeT) -> ModuleType:
    file_path = Path(str(module_file_path)).resolve()
    module_name = file_path.stem

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is None:
        raise ImportError(f"Cannot find module {module_name} at {file_path}")

    module = importlib.util.module_from_spec(spec)
    loader = spec.loader
    if loader is None or not hasattr(loader, "exec_module"):
        raise ImportError(f"Cannot load module {module_name} from {file_path}")

    loader.exec_module(module)
    sys.modules[module_name] = module

    return module



import logging
import os
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from urllib3.util import parse_url

from vastdb.bench.perf_bench.common.types import StrEnum

_MY_DIR = Path(__file__).parent

# VastDB details
VASTDB_ENDPOINT = os.getenv("VASTDB_ENDPOINT", "")
VASTDB_BUCKET_NAME = os.getenv("VASTDB_BUCKET_NAME", "")
VASTDB_TEST_SCHEMA_NAME = os.getenv("VASTDB_TEST_SCHEMA_NAME", "")
VASTDB_TEST_TABLE_NAME = os.getenv("VASTDB_TEST_TABLE_NAME", "")

# Regular S3 details
S3_BUCKET_NAME = "my-s3-bucket"
DEFAULT_S3_SSL_PORT = int(os.getenv("DEFAULT_S3_SSL_PORT", 443))
if DEFAULT_S3_ENDPOINT_URL := os.getenv("AWS_S3_ENDPOINT_URL", ""):
    _parsed = parse_url(DEFAULT_S3_ENDPOINT_URL)
    DEFAULT_S3_HOST = str(_parsed.host)
    DEFAULT_S3_PORT = int(_parsed.port or 80)
else:
    DEFAULT_S3_HOST = os.getenv("DEFAULT_S3_HOST", "1.1.1.1")
    DEFAULT_S3_PORT = int(os.getenv("DEFAULT_S3_PORT", 80))
    DEFAULT_S3_ENDPOINT_URL = f"http://{DEFAULT_S3_HOST}:{DEFAULT_S3_PORT}"

# Paths
DEFAULT_RESULTS_DIR = Path(__file__).parent.parent / "benchmark_results"
NFS_DS_PATH = f"/mnt/data/{VASTDB_BUCKET_NAME}/data"
LOCAL_FS_DS_PATH = _MY_DIR.parent / "dataset" / "test_dataset"
S3_DS_PATH = f"s3://{VASTDB_BUCKET_NAME}/data"

# Access keys
DEFAULT_ACCESS_KEY = os.getenv("AWS_ACCESS_KEY_ID", "some_access_key")
DEFAULT_SECRET_KEY = os.getenv("AWS_SECRET_ACCESS_KEY", "some_secret_key")

# Dataset start/end dates
DEFAULT_START_T = pd.Timestamp("20180101")
DEFAULT_END_T = pd.Timestamp("20200201")

# Arrow related constants
DEFAULT_ROW_GROUP_SIZE = 1024 * 1024
DEFAULT_ARROW_KWARGS = {
    "batch_size": (DEFAULT_ARROW_BATCH_SIZE := 786432),
    "batch_readahead": (DEFAULT_ARROW_BATCH_READAHEAD := 12),
    "fragment_readahead": (DEFAULT_ARROW_BATCH_FRAGMENT_READAHEAD := 4),
}


class ParquetCompression(StrEnum):
    NONE = "NONE"
    SNAPPY = "SNAPPY"
    GZIP = "GZIP"
    # LZO = "LZO"
    BROTLI = "BROTLI"
    LZ4 = "LZ4"
    ZSTD = "ZSTD"


DFAULT_PARQUET_COMPRESSION = ParquetCompression.LZ4


class LogLevel(StrEnum):
    CRITICAL = "CRITICAL"
    FATAL = "FATAL"
    ERROR = "ERROR"
    WARN = "WARNING"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"
    NOTSET = "NOTSET"

    def to_int(self) -> int:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        return logging._nameToLevel[self]  # noqa: SLF001


@dataclass(frozen=True)
class VastConnDetails:
    """VAST Config."""

    access: str = DEFAULT_ACCESS_KEY
    secret: str = DEFAULT_SECRET_KEY
    vastdb_bucket: str = VASTDB_BUCKET_NAME
    vastdb_endpoint: str = VASTDB_ENDPOINT
    vastdb_ssl_verify: bool = True
    vastdb_schema: str = VASTDB_TEST_SCHEMA_NAME
    vastdb_table: str = VASTDB_TEST_TABLE_NAME
    s3_host: str = DEFAULT_S3_HOST
    s3_bucket: str = VASTDB_BUCKET_NAME
    s3_port: int = DEFAULT_S3_PORT
    s3_ssl_port: int = DEFAULT_S3_SSL_PORT



"""VAST database session.

It should be used to interact with a specific VAST cluster.
For more details see:
- [Virtual IP pool configured with DNS service](https://support.vastdata.com/s/topic/0TOV40000000FThOAM/configuring-network-access-v50)
- [S3 access & secret keys on VAST cluster](https://support.vastdata.com/s/article/UUID-4d2e7e23-b2fb-7900-d98f-96c31a499626)
- [Tabular identity policy with the proper permissions](https://support.vastdata.com/s/article/UUID-14322b60-d6a2-89ac-3df0-3dfbb6974182)
"""

import os
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .config import BackoffConfig


class Session:
    """VAST database session."""

    def __init__(self, access=None, secret=None, endpoint=None,
                 *,
                 ssl_verify=True,
                 timeout=None,
                 backoff_config: Optional["BackoffConfig"] = None):
        """Connect to a VAST Database endpoint, using specified credentials."""
        from . import _internal, features

        if access is None:
            access = os.environ['AWS_ACCESS_KEY_ID']
        if secret is None:
            secret = os.environ['AWS_SECRET_ACCESS_KEY']
        if endpoint is None:
            endpoint = os.environ['AWS_S3_ENDPOINT_URL']

        self.api = _internal.VastdbApi(
            endpoint=endpoint,
            access_key=access,
            secret_key=secret,
            ssl_verify=ssl_verify,
            timeout=timeout,
            backoff_config=backoff_config)
        self.features = features.Features(self.api.vast_version)

    def __repr__(self):
        """Don't show the secret key."""
        return f'{self.__class__.__name__}(endpoint={self.api.url}, access={self.api.access_key})'

    def transaction(self):
        """Create a non-initialized transaction object.

        It should be used as a context manager:

            with session.transaction() as tx:
                tx.bucket("bucket").create_schema("schema")
        """
        from . import transaction
        return transaction.Transaction(self)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# A union is a complex type with children in Field
# By default ids in the type vector refer to the offsets in the children
# optionally typeIds provides an indirection between the child offset and the type id
# for each child `typeIds[offset]` is the id used in the type vector
class Union(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Union()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsUnion(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Union
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Union
    def Mode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # Union
    def TypeIds(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Union
    def TypeIdsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Union
    def TypeIdsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Union
    def TypeIdsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def Start(builder): builder.StartObject(2)
def UnionStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddMode(builder, mode): builder.PrependInt16Slot(0, mode, 0)
def UnionAddMode(builder, mode):
    """This method is deprecated. Please switch to AddMode."""
    return AddMode(builder, mode)
def AddTypeIds(builder, typeIds): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(typeIds), 0)
def UnionAddTypeIds(builder, typeIds):
    """This method is deprecated. Please switch to AddTypeIds."""
    return AddTypeIds(builder, typeIds)
def StartTypeIdsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def UnionStartTypeIdsVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartTypeIdsVector(builder, numElems)
def End(builder): return builder.EndObject()
def UnionEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class CreateViewRequest(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = CreateViewRequest()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCreateViewRequest(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # CreateViewRequest
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CreateViewRequest
    def ViewMetadataArrowBuffer(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # CreateViewRequest
    def ViewMetadataArrowBufferAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # CreateViewRequest
    def ViewMetadataArrowBufferLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # CreateViewRequest
    def ViewMetadataArrowBufferIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # CreateViewRequest
    def ViewDataArrowSchema(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # CreateViewRequest
    def ViewDataArrowSchemaAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # CreateViewRequest
    def ViewDataArrowSchemaLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # CreateViewRequest
    def ViewDataArrowSchemaIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def Start(builder): builder.StartObject(2)
def CreateViewRequestStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddViewMetadataArrowBuffer(builder, viewMetadataArrowBuffer): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(viewMetadataArrowBuffer), 0)
def CreateViewRequestAddViewMetadataArrowBuffer(builder, viewMetadataArrowBuffer):
    """This method is deprecated. Please switch to AddViewMetadataArrowBuffer."""
    return AddViewMetadataArrowBuffer(builder, viewMetadataArrowBuffer)
def StartViewMetadataArrowBufferVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def CreateViewRequestStartViewMetadataArrowBufferVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartViewMetadataArrowBufferVector(builder, numElems)
def AddViewDataArrowSchema(builder, viewDataArrowSchema): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(viewDataArrowSchema), 0)
def CreateViewRequestAddViewDataArrowSchema(builder, viewDataArrowSchema):
    """This method is deprecated. Please switch to AddViewDataArrowSchema."""
    return AddViewDataArrowSchema(builder, viewDataArrowSchema)
def StartViewDataArrowSchemaVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def CreateViewRequestStartViewDataArrowSchemaVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartViewDataArrowSchemaVector(builder, numElems)
def End(builder): return builder.EndObject()
def CreateViewRequestEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# ruff: noqa: PLW2901, C901

import datetime as dt
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Sequence

import pandas as pd

from vastdb.bench.perf_bench.common.constants import (
    DEFAULT_RESULTS_DIR,
)
from vastdb.bench.perf_bench.common.log_utils import get_logger
from vastdb.bench.perf_bench.common.types import PathLikeT
from vastdb.bench.perf_bench.orchestrate import results_helpers as bc
from vastdb.bench.perf_bench.orchestrate.bench_spec import BenchResult
from vastdb.bench.perf_bench.orchestrate.scenario import BenchScenario

LOG = get_logger(__name__)


@dataclass(frozen=True)
class RunnableBenchSession:
    bench_name: str
    scenarios: Sequence[BenchScenario]
    runs_per_bench: int
    parallelism: int
    results_base_dir: Optional[PathLikeT]


def run_scenarios(
    scenarios: Sequence[BenchScenario],
    runs_per_bench: int = 3,
    parallelism: int = 1,
    results_base_dir: Optional[str] = None,
):
    results_path = (
        Path(results_base_dir or DEFAULT_RESULTS_DIR)
        / f"results_{dt.datetime.now().strftime('%Y%m%d_%H%M%S')}_par{parallelism}.csv"
    )

    if (runs_per_bench := int(runs_per_bench)) < 1:
        raise ValueError(f"runs_per_bench must be >= 1: {runs_per_bench=}")
    if (parallelism := int(parallelism)) < 1:
        raise ValueError(f"parallelism must be >= 1: {parallelism=}")

    LOG.info("Will run '%d' scenarios", len(scenarios))

    # noinspection PyShadowingNames
    def worker_init():
        import numpy as np  # noqa: F401
        import pandas as pd  # noqa: F401
        import pyarrow as pa  # noqa: F401

        from vastdb.bench.perf_bench.common.utils import get_logger

        get_logger(__name__).info("Warmed up.")

    results: List[BenchResult] = []
    for i, scen in enumerate(scenarios):
        LOG.info(f"Running scenario: {scen.key} [{i + 1} of {len(scenarios)}]")
        scen.run(
            n_runs=runs_per_bench,
            discard_first_run=True,
            parallelism=parallelism,
            workers_init=worker_init,
        )
        results.extend(scen.results or ())
        bc.save_results(results, results_path=results_path)

    # Save the results from all workers and print stats
    bc.save_results(results, results_path=results_path)
    results_stats_df = bc.calculate_aggregate_stats(results=results)
    LOG.info(f"\n{results_stats_df.to_string()}")

    # Verify the results written
    results_df = bc.results_to_df(results)
    results_df2 = pd.read_csv(results_path).astype(results_df.dtypes)
    pd.testing.assert_frame_equal(results_df, results_df2)



import logging
from tempfile import NamedTemporaryFile

import pyarrow as pa
import pyarrow.parquet as pq
import pytest

from vastdb import util
from vastdb.errors import ImportFilesError, InternalServerError, InvalidArgument

log = logging.getLogger(__name__)


def test_parallel_imports(session, clean_bucket_name, s3):
    num_rows = 1000
    num_files = 53
    ds = {'num': [i for i in range(num_rows)]}
    files = []
    table = pa.Table.from_pydict(ds)
    with NamedTemporaryFile() as f:
        pq.write_table(table, f.name)
        s3.put_object(Bucket=clean_bucket_name, Key='prq0', Body=f)
        files.append(f'/{clean_bucket_name}/prq0')

    for i in range(1, num_files):
        copy_source = {
            'Bucket': clean_bucket_name,
            'Key': 'prq0'
        }
        s3.copy(copy_source, clean_bucket_name, f'prq{i}')
        files.append(f'/{clean_bucket_name}/prq{i}')

    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        s = b.create_schema('s1')
        t = s.create_table('t1', pa.schema([('num', pa.int64())]))
        with pytest.raises(InternalServerError):
            t.create_imports_table()
        log.info("Starting import of %d files", num_files)
        t.import_files(files)
        arrow_table = t.select(columns=['num']).read_all()
        assert arrow_table.num_rows == num_rows * num_files
        arrow_table = t.select(columns=['num'], predicate=t['num'] == 100).read_all()
        assert arrow_table.num_rows == num_files
        import_table = t.imports_table()
        # checking all imports are on the imports table:
        objects_name = import_table.select(columns=["ObjectName"]).read_all()
        objects_name = objects_name.to_pydict()
        object_names = set(objects_name['ObjectName'])
        prefix = 'prq'
        numbers = set(range(53))
        assert all(name.startswith(prefix) for name in object_names)
        numbers.issubset(int(name.replace(prefix, '')) for name in object_names)
        assert len(object_names) == len(objects_name['ObjectName'])


def test_create_table_from_files(session, clean_bucket_name, s3):
    datasets = [
        {'num': [0],
         'varch': ['z']},
        {'num': [1, 2, 3, 4, 5],
         'varch': ['a', 'b', 'c', 'd', 'e']},
        {'num': [1, 2, 3, 4, 5],
         'bool': [True, False, None, None, False],
         'varch': ['a', 'b', 'c', 'd', 'e']},
        {'num': [1, 2],
         'bool': [True, True]},
        {'varch': ['a', 'b', 'c'],
         'mismatch': [1, 2, 3]}
    ]
    for i, ds in enumerate(datasets):
        table = pa.Table.from_pydict(ds)
        with NamedTemporaryFile() as f:
            pq.write_table(table, f.name)
            s3.put_object(Bucket=clean_bucket_name, Key=f'prq{i}', Body=f)

    same_schema_files = [f'/{clean_bucket_name}/prq{i}' for i in range(2)]
    contained_schema_files = [f'/{clean_bucket_name}/prq{i}' for i in range(4)]
    different_schema_files = [f'/{clean_bucket_name}/prq{i}' for i in range(5)]

    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        s = b.create_schema('s1')
        t = util.create_table_from_files(s, 't1', contained_schema_files)
        assert len(t.arrow_schema) == 3
        assert t.arrow_schema == pa.schema([('num', pa.int64()), ('bool', pa.bool_()), ('varch', pa.string())])

        with pytest.raises(InvalidArgument):
            util.create_table_from_files(s, 't2', different_schema_files)

        with pytest.raises(InvalidArgument):
            util.create_table_from_files(s, 't2', contained_schema_files, schema_merge_func=util.strict_schema_merge)

        util.create_table_from_files(s, 't2', different_schema_files, schema_merge_func=util.union_schema_merge)
        util.create_table_from_files(s, 't3', same_schema_files, schema_merge_func=util.strict_schema_merge)


def test_import_name_mismatch_error(session, clean_bucket_name, s3):
    ds = {'varch': ['a', 'b', 'c'],
          'invalid_column_name': [1, 2, 3]}
    prq_name = 'name_mismatch.parquet'
    table = pa.Table.from_pydict(ds)
    with NamedTemporaryFile() as f:
        pq.write_table(table, f.name)
        s3.put_object(Bucket=clean_bucket_name, Key=prq_name, Body=f)

    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        s = b.create_schema('s1')
        t = s.create_table('t1', pa.schema([('varch', pa.string()), ('num', pa.int64())]))
        with pytest.raises(ImportFilesError) as exc:
            t.import_files([f'/{clean_bucket_name}/{prq_name}'])
        assert exc.value.error_dict['object_name'] == prq_name
        assert exc.value.error_dict['res'] == 'TabularMismatchColumnName'
        assert 'invalid_column_name' in exc.value.error_dict['err_msg']


def test_import_type_mismatch_error(session, clean_bucket_name, s3):
    ds = {'varch': ['a', 'b', 'c'],
          'num_type_mismatch': [1, 2, 3]}
    prq_name = 'type_mismatch.parquet'
    table = pa.Table.from_pydict(ds)
    with NamedTemporaryFile() as f:
        pq.write_table(table, f.name)
        s3.put_object(Bucket=clean_bucket_name, Key=prq_name, Body=f)

    with session.transaction() as tx:
        b = tx.bucket(clean_bucket_name)
        s = b.create_schema('s1')
        t = s.create_table('t1', pa.schema([('varch', pa.string()), ('num_type_mismatch', pa.bool_())]))
        with pytest.raises(ImportFilesError) as exc:
            t.import_files([f'/{clean_bucket_name}/{prq_name}'])
        assert exc.value.error_dict['object_name'] == prq_name
        assert exc.value.error_dict['res'] == 'TabularMismatchColumnType'
        assert 'num_type_mismatch' in exc.value.error_dict['err_msg']



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Join between two tables
class Join(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Join()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsJoin(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Join
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # An identifiier for the relation. The identifier should be unique over the
    # entire plan. Optional.
    # Join
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.RelId import RelId
            obj = RelId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Left relation
    # Join
    def Left(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.Relation import Relation
            obj = Relation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Right relation
    # Join
    def Right(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.Relation import Relation
            obj = Relation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # The expression which will be evaluated against rows from each
    # input to determine whether they should be included in the
    # join relation's output.
    # Join
    def OnExpression(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # The kind of join to use.
    # Join
    def JoinKind(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(5)
def JoinStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddId(builder, id): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def JoinAddId(builder, id):
    """This method is deprecated. Please switch to AddId."""
    return AddId(builder, id)
def AddLeft(builder, left): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(left), 0)
def JoinAddLeft(builder, left):
    """This method is deprecated. Please switch to AddLeft."""
    return AddLeft(builder, left)
def AddRight(builder, right): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(right), 0)
def JoinAddRight(builder, right):
    """This method is deprecated. Please switch to AddRight."""
    return AddRight(builder, right)
def AddOnExpression(builder, onExpression): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(onExpression), 0)
def JoinAddOnExpression(builder, onExpression):
    """This method is deprecated. Please switch to AddOnExpression."""
    return AddOnExpression(builder, onExpression)
def AddJoinKind(builder, joinKind): builder.PrependUint8Slot(4, joinKind, 0)
def JoinAddJoinKind(builder, joinKind):
    """This method is deprecated. Please switch to AddJoinKind."""
    return AddJoinKind(builder, joinKind)
def End(builder): return builder.EndObject()
def JoinEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Literal relation
class LiteralRelation(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = LiteralRelation()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsLiteralRelation(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # LiteralRelation
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # An identifiier for the relation. The identifier should be unique over the
    # entire plan. Optional.
    # LiteralRelation
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.RelId import RelId
            obj = RelId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # The columns of this literal relation.
    # LiteralRelation
    def Columns(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vast_flatbuf.org.apache.arrow.computeir.flatbuf.LiteralColumn import LiteralColumn
            obj = LiteralColumn()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # LiteralRelation
    def ColumnsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # LiteralRelation
    def ColumnsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def Start(builder): builder.StartObject(2)
def LiteralRelationStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddId(builder, id): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def LiteralRelationAddId(builder, id):
    """This method is deprecated. Please switch to AddId."""
    return AddId(builder, id)
def AddColumns(builder, columns): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(columns), 0)
def LiteralRelationAddColumns(builder, columns):
    """This method is deprecated. Please switch to AddColumns."""
    return AddColumns(builder, columns)
def StartColumnsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def LiteralRelationStartColumnsVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartColumnsVector(builder, numElems)
def End(builder): return builder.EndObject()
def LiteralRelationEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

class IntervalLiteralImpl(object):
    NONE = 0
    IntervalLiteralMonths = 1
    IntervalLiteralDaysMilliseconds = 2




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# ----------------------------------------------------------------------
# Data structures for describing a table row batch (a collection of
# equal-length Arrow arrays)
# Metadata about a field at some level of a nested type tree (but not
# its children).
#
# For example, a List<Int16> with values `[[1, 2, 3], null, [4], [5, 6], null]`
# would have {length: 5, null_count: 2} for its List node, and {length: 6,
# null_count: 0} for its Int16 node, as separate FieldNode structs
class FieldNode(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 16

    # FieldNode
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The number of value slots in the Arrow array at this level of a nested
    # tree
    # FieldNode
    def Length(self): return self._tab.Get(flatbuffers.number_types.Int64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # The number of observed nulls. Fields with null_count == 0 may choose not
    # to write their physical validity bitmap out as a materialized buffer,
    # instead setting the length of the bitmap buffer to 0.
    # FieldNode
    def NullCount(self): return self._tab.Get(flatbuffers.number_types.Int64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateFieldNode(builder, length, nullCount):
    builder.Prep(8, 16)
    builder.PrependInt64(nullCount)
    builder.PrependInt64(length)
    return builder.Offset()



from pathlib import Path
from typing import Annotated, List, Optional

import pandas as pd
import typer

from vastdb.bench.perf_bench.common.constants import (
    DEFAULT_END_T,
    DEFAULT_RESULTS_DIR,
    DEFAULT_START_T,
    DFAULT_PARQUET_COMPRESSION,
    LOCAL_FS_DS_PATH,
    LogLevel,
    ParquetCompression,
)
from vastdb.bench.perf_bench.common.log_utils import (
    get_logger,
    set_log_file,
    set_log_level,
)
from vastdb.bench.perf_bench.common.utils import getenv_flag, load_module_from_path
from vastdb.bench.perf_bench.dataset.generate_secmaster import (
    SM_PATH,
    generate_secmaster,
)
from vastdb.bench.perf_bench.dataset.generate_stocks_dataset import (
    generate_concurrent_synthetic_stock_1m_bars,
)
from vastdb.bench.perf_bench.orchestrate.scenario import BenchScenario
from vastdb.bench.perf_bench.run import run_scenarios

app = typer.Typer(pretty_exceptions_enable=getenv_flag("TYPER_PRETTY_EXCEPTIONS"))

_MY_DIR = Path(__file__).parent


# noinspection PyUnusedLocal
@app.callback()
def cli_common(
    ctx: typer.Context,
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose",
            is_flag=True,
        ),
    ] = False,
    log_level: Annotated[
        Optional[LogLevel],
        typer.Option(
            "--log-level",
            case_sensitive=False,
        ),
    ] = None,
    log_file: Annotated[
        Optional[Path],
        typer.Option(
            "--log-file",
            writable=True,
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ] = None,
):
    if verbose:
        log_level = LogLevel.DEBUG
    if log_level:
        set_log_level(log_level)
    if log_file:
        set_log_file(log_file)
    get_logger(__name__).info("CLI common setup done.")


def _positive_int(value: str) -> int:
    i_value = int(value)
    if i_value <= 0:
        raise typer.BadParameter(f"Must be a positive integer: {value}.")
    return i_value


# noinspection PyUnusedLocal
@app.command()
def run_bench(
    ctx: typer.Context,
    bench_name: Annotated[
        str,
        typer.Option(
            "--bench-name",
        ),
    ],
    parallelism: Annotated[
        List[int],
        typer.Option(
            "--parallelism",
            callback=lambda par: [_positive_int(p) for p in par],
        ),
    ],
    runs_per_bench: Annotated[
        int,
        typer.Option(
            "--runs-per-bench",
            callback=_positive_int,
        ),
    ] = 3,
    bench_generator_path: Annotated[
        Path,
        typer.Option(
            "--bench-generator-path",
            readable=True,
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ] = _MY_DIR / "bench_repo" / "mega_combo.py",
    results_base_dir: Annotated[
        Path,
        typer.Option(
            "--log-file",
            writable=True,
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
        ),
    ] = DEFAULT_RESULTS_DIR,
):
    if not (bench_name := bench_name.strip()):
        raise typer.BadParameter("Bench name must be non-empty.")

    mod = load_module_from_path(bench_generator_path)
    scenarios: List[BenchScenario] = mod.build_scenarios(base_key=bench_name)
    for para in parallelism:
        run_scenarios(
            scenarios=scenarios,
            runs_per_bench=runs_per_bench,
            parallelism=para,
            results_base_dir=str(results_base_dir),
        )


# noinspection PyUnusedLocal
@app.command()
def build_secmaster(
    ctx: typer.Context,
):
    generate_secmaster()


# noinspection PyUnusedLocal
@app.command()
def build_dataset(
    ctx: typer.Context,
    start_date: Annotated[
        str,
        typer.Option(
            "--start-date",
            help="Start date for the dataset.",
            callback=lambda d: pd.Timestamp(d).normalize(),
        ),
    ] = DEFAULT_START_T.strftime("%Y%m%d"),
    end_date: Annotated[
        str,
        typer.Option(
            "--end-date",
            help="Start date for the dataset.",
            callback=lambda d: pd.Timestamp(d).normalize(),
        ),
    ] = DEFAULT_END_T.strftime("%Y%m%d"),
    output_dir: Annotated[
        Path,
        typer.Option(
            "--output-dir",
            writable=True,
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
        ),
    ] = LOCAL_FS_DS_PATH,
    parallelism: Annotated[
        int,
        typer.Option(
            "--parallelism",
            callback=_positive_int,
        ),
    ] = 6,
    row_group_size: Annotated[
        int,
        typer.Option(
            "--row-group-size",
            callback=_positive_int,
            help=(
                "Row group size for the dataset, some common values are: 64 * 1024, 128 * 1024, 256"
                " * 1024, 512 * 1024,1024 * 1024, 1.5 * 1024 * 1024."
            ),
        ),
    ] = 256 * 1024,
    compression: Annotated[
        ParquetCompression,
        typer.Option(
            "--compression",
            help="Parquet compression algorithm.",
        ),
    ] = DFAULT_PARQUET_COMPRESSION,
):
    if row_group_size < 1024:
        raise typer.BadParameter("Row group size must be at least 1024.")
    if parallelism < 1:
        raise typer.BadParameter("Parallelism must be at least 1.")
    if start_date > end_date:
        raise typer.BadParameter("Start date must be before the end date.")
    if not SM_PATH.is_file():
        generate_secmaster()
    generate_concurrent_synthetic_stock_1m_bars(
        from_t=start_date,
        to_t=end_date,
        output_dir=output_dir,
        num_workers=parallelism,
        row_group_size=row_group_size,
        compression=compression,
    )


if __name__ == "__main__":
    # Set the metadata only if we execute the main (not on just importing this module)
    app()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class DictionaryEncoding(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DictionaryEncoding()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsDictionaryEncoding(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # DictionaryEncoding
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The known dictionary id in the application where this data is used. In
    # the file or streaming formats, the dictionary ids are found in the
    # DictionaryBatch messages
    # DictionaryEncoding
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # The dictionary indices are constrained to be non-negative integers. If
    # this field is null, the indices must be signed int32. To maximize
    # cross-language compatibility and performance, implementations are
    # recommended to prefer signed integer types over unsigned integer types
    # and to avoid uint64 indices unless they are required by an application.
    # DictionaryEncoding
    def IndexType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vast_flatbuf.org.apache.arrow.flatbuf.Int import Int
            obj = Int()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # By default, dictionaries are not ordered, or the order does not have
    # semantic meaning. In some statistical, applications, dictionary-encoding
    # is used to represent ordered categorical data, and we provide a way to
    # preserve that metadata here
    # DictionaryEncoding
    def IsOrdered(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # DictionaryEncoding
    def DictionaryKind(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(4)
def DictionaryEncodingStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddId(builder, id): builder.PrependInt64Slot(0, id, 0)
def DictionaryEncodingAddId(builder, id):
    """This method is deprecated. Please switch to AddId."""
    return AddId(builder, id)
def AddIndexType(builder, indexType): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(indexType), 0)
def DictionaryEncodingAddIndexType(builder, indexType):
    """This method is deprecated. Please switch to AddIndexType."""
    return AddIndexType(builder, indexType)
def AddIsOrdered(builder, isOrdered): builder.PrependBoolSlot(2, isOrdered, 0)
def DictionaryEncodingAddIsOrdered(builder, isOrdered):
    """This method is deprecated. Please switch to AddIsOrdered."""
    return AddIsOrdered(builder, isOrdered)
def AddDictionaryKind(builder, dictionaryKind): builder.PrependInt16Slot(3, dictionaryKind, 0)
def DictionaryEncodingAddDictionaryKind(builder, dictionaryKind):
    """This method is deprecated. Please switch to AddDictionaryKind."""
    return AddDictionaryKind(builder, dictionaryKind)
def End(builder): return builder.EndObject()
def DictionaryEncodingEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


"""VAST Database schema (a container of tables).

VAST S3 buckets can be used to create Database schemas and tables.
It is possible to list and access VAST snapshots generated over a bucket.
"""

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterable, List, Optional

import pyarrow as pa

from . import bucket, errors, schema, table

if TYPE_CHECKING:
    from .table import Table


log = logging.getLogger(__name__)


@dataclass
class Schema:
    """VAST Schema."""

    name: str
    bucket: "bucket.Bucket"

    @property
    def tx(self):
        """VAST transaction used for this schema."""
        return self.bucket.tx

    def _subschema_full_name(self, name: str) -> str:
        return f"{self.name}/{name}" if self.name else name

    def create_schema(self, name: str, fail_if_exists=True) -> "Schema":
        """Create a new schema (a container of tables) under this schema."""
        if current := self.schema(name, fail_if_missing=False):
            if fail_if_exists:
                raise errors.SchemaExists(self.bucket.name, name)
            else:
                return current
        full_name = self._subschema_full_name(name)
        self.tx._rpc.api.create_schema(self.bucket.name, full_name, txid=self.tx.txid)
        log.info("Created schema: %s", full_name)
        return self.schema(name)  # type: ignore[return-value]

    def schema(self, name: str, fail_if_missing=True) -> Optional["Schema"]:
        """Get a specific schema (a container of tables) under this schema."""
        _bucket_name, schemas, _next_key, _is_truncated, _ = \
            self.tx._rpc.api.list_schemas(bucket=self.bucket.name, schema=self.name, next_key=0, txid=self.tx.txid,
                                          name_prefix=name, exact_match=True, max_keys=1)
        names = [name for name, *_ in schemas]
        log.debug("Found schemas: %s", names)
        if not names:
            if fail_if_missing:
                raise errors.MissingSchema(self.bucket.name, self._subschema_full_name(name))
            else:
                return None

        assert len(names) == 1, f"Expected to receive only a single schema, but got {len(schemas)}: ({schemas})"
        return schema.Schema(name=self._subschema_full_name(names[0]), bucket=self.bucket)

    def schemas(self, batch_size=None) -> Iterable["Schema"]:
        """List child schemas."""
        next_key = 0
        if not batch_size:
            batch_size = 1000
        result: List["Schema"] = []
        while True:
            _bucket_name, curr_schemas, next_key, is_truncated, _ = \
                self.tx._rpc.api.list_schemas(bucket=self.bucket.name, schema=self.name, next_key=next_key, max_keys=batch_size, txid=self.tx.txid)
            result.extend(schema.Schema(name=self._subschema_full_name(name), bucket=self.bucket) for name, *_ in curr_schemas)
            if not is_truncated:
                break
        return result

    def create_table(self, table_name: str, columns: pa.Schema, fail_if_exists=True, use_external_row_ids_allocation=False) -> "Table":
        """Create a new table under this schema.

        A virtual `vastdb_rowid` column (of `int64` type) can be created to access and filter by internal VAST row IDs.
        See https://support.vastdata.com/s/article/UUID-48d0a8cf-5786-5ef3-3fa3-9c64e63a0967 for more details.
        """
        if current := self.table(table_name, fail_if_missing=False):
            if fail_if_exists:
                raise errors.TableExists(self.bucket.name, self.name, table_name)
            else:
                return current
        if use_external_row_ids_allocation:
            self.tx._rpc.features.check_external_row_ids_allocation()

        self.tx._rpc.api.create_table(self.bucket.name, self.name, table_name, columns, txid=self.tx.txid,
                                      use_external_row_ids_allocation=use_external_row_ids_allocation)
        log.info("Created table: %s", table_name)
        return self.table(table_name)  # type: ignore[return-value]

    def table(self, name: str, fail_if_missing=True) -> Optional["table.Table"]:
        """Get a specific table under this schema."""
        t = self.tables(table_name=name)
        if not t:
            if fail_if_missing:
                raise errors.MissingTable(self.bucket.name, self.name, name)
            else:
                return None
        assert len(t) == 1, f"Expected to receive only a single table, but got: {len(t)}. tables: {t}"
        log.debug("Found table: %s", t[0])
        return t[0]

    def _iter_tables(self, table_name=None):
        next_key = 0
        name_prefix = table_name if table_name else ""
        exact_match = bool(table_name)
        while True:
            _bucket_name, _schema_name, curr_tables, next_key, is_truncated, _ = \
                self.tx._rpc.api.list_tables(
                    bucket=self.bucket.name, schema=self.name, next_key=next_key, txid=self.tx.txid,
                    exact_match=exact_match, name_prefix=name_prefix, include_list_stats=exact_match)
            if not curr_tables:
                break
            yield from curr_tables
            if not is_truncated:
                break

    def tables(self, table_name=None) -> List["Table"]:
        """List all tables under this schema."""
        return [
            _parse_table_info(table_info, self)
            for table_info in self._iter_tables(table_name=table_name)
        ]

    def tablenames(self) -> List[str]:
        """List all table names under this schema."""
        return [table_info.name for table_info in self._iter_tables()]

    def drop(self) -> None:
        """Delete this schema."""
        self.tx._rpc.api.drop_schema(self.bucket.name, self.name, txid=self.tx.txid)
        log.info("Dropped schema: %s", self.name)

    def rename(self, new_name) -> None:
        """Rename this schema."""
        self.tx._rpc.api.alter_schema(self.bucket.name, self.name, txid=self.tx.txid, new_name=new_name)
        log.info("Renamed schema: %s to %s", self.name, new_name)
        self.name = new_name


def _parse_table_info(table_info, schema: "schema.Schema"):
    return table.Table(name=table_info.name, schema=schema, handle=int(table_info.handle), _imports_table=False)
